import Koa from "koa"
import serve from "koa-static"
import http from "http"
import webpack from "webpack"
import path from "path"
import {devMiddleware, hotMiddleware} from "koa-webpack-middleware"
import VirtualModulesPlugin from "webpack-virtual-modules"
import MemoryFS from "memory-fs"
import realFs from "fs"

const fs = new MemoryFS()

const externalDeps = {
  "react-dom": `commonjs react-dom`,
  "react": `commonjs react`,
  "classnames": `commonjs classnames`,
  "core-js": `commonjs core-js`,
  "history": `commonjs history`,
  "react-redux": `commonjs react-redux`,
  "redux": `commonjs redux`,
}

const sandboxVirtualModules = new VirtualModulesPlugin()

const loadTmpModules = () => {
  const cwd = process.cwd()
  try {
    const tmpContent = realFs.readFileSync(path.resolve(cwd, `src/tmp-content.json`), `utf8`)
    sandboxVirtualModules.writeModule(`node_modules/weblium/block-content.json`, tmpContent)
  } catch (e) {
    sandboxVirtualModules.writeModule(`node_modules/weblium/block-content.json`, '{}')
  }
  try {
    const tmpModifier = realFs.readFileSync(path.resolve(cwd, `src/tmp-modifier.json`), `utf8`)
    sandboxVirtualModules.writeModule(`node_modules/weblium/block-modifier.json`, tmpModifier)
  } catch (e) {
    sandboxVirtualModules.writeModule(`node_modules/weblium/block-modifier.json`, '{}')
  }
  try {
    const blockMap = fs.readFileSync(`/block.js.map`).toString(`utf8`)
    sandboxVirtualModules.writeModule(`node_modules/block.js.map`, blockMap)
  } catch (e) {
    console.log(e)
  }

  try {
    const themeBlocks = process.env.WM_SANDBOX_BLOCKS || realFs.readFileSync(path.resolve(cwd, 'theme-blocks.txt'), 'utf8')
    sandboxVirtualModules.writeModule(
      'node_modules/weblium/theme-blocks',
      `module.exports = (function() {return ${JSON.stringify(themeBlocks)}.split('\\n').filter(Boolean)})()`)
  } catch (e) {
    sandboxVirtualModules.writeModule('node_modules/weblium/theme-blocks', 'module.exports = []')
  }

  try {
    fs.statSync(`/block.js`).isFile()
    const blockModule = fs.readFileSync(`/block.js`).toString(`utf8`)
    sandboxVirtualModules.writeModule(`node_modules/block.js`, blockModule)
  } catch (e) {
    console.log(e)
  }

}
const getSandboxConfig = async (type = '') => {
  const cwd = process.cwd()
  const editorPath = type.match(/editor|theme/)
    ? path.resolve(__dirname, `../../sandbox-full.js`)
    : path.resolve(__dirname, `../../sandbox-mini.js`)

  return {
    entry: [
      ...(type.match(/editor|theme/) ? [] : [`babel-polyfill`]),
      `./app/sandbox/client/index.js`,
      `webpack-hot-middleware/client?path=/__webpack_hmr&timeout=20000`,
    ],
    context: path.join(__dirname, `../../..`),
    resolve: {
      alias: {
        'weblium/editor': editorPath,
      },
      modules: [`node_modules`],
    },
    mode: 'development',
    module: {
      rules: [
        {
          test: /\.(js|jsx)$/,
          loader: `babel-loader`,
          exclude: /node_modules[/\\].+[/\\]node_modules|sandbox-full|sandbox-mini/,
          options: {
            presets: [
              require.resolve(`babel-preset-env`),
              require.resolve(`babel-preset-react`),
            ],
            plugins: [
              require.resolve(`babel-plugin-transform-decorators-legacy`),
              require.resolve(`babel-plugin-transform-class-properties`),
              require.resolve(`babel-plugin-transform-object-rest-spread`),
            ],
          },
        },
        {
          test: /\.s?css$/,
          use: [
            {
              loader: `style-loader`,
              options: {
                singleton: true,
              },
            },
            {
              loader: `css-loader`,
              options: {
                sourceMap: true,
                importLoaders: 1,
              },
            },
            {
              loader: `postcss-loader`, // translates CSS into CommonJS
              options: {
                plugins: () => [
                  require('postcss-nested'),
                ],
              },
            },
          ],
        },
      ],
    },
    plugins: [
      sandboxVirtualModules,
      new webpack.HotModuleReplacementPlugin(),
      new webpack.ProvidePlugin({
        React: `react`,
        PropTypes: `prop-types`,
        _: `lodash/fp`,
        classNames: `classnames`,
      }),
      new webpack.DefinePlugin({
        __DEV__: false,
        __PROD__: true,
        "process.env.NODE_ENV": JSON.stringify(process.env.NODE_ENV || 'development'),
        __MODE__: JSON.stringify(type),
      }),
    ],
  }
}

const getBlockConfig = async (type = '') => {
  const cwd = process.cwd()
  const styleLoader = type.match(/theme/)
    ? {
      test: /\.s?css$/,
      exclude: /node_modules/,
      use: [
        {
          loader: `style-loader`,
          options: {
            singleton: false,
          },
        }, {
          loader: 'css-loader',
          options: {
            importLoaders: 2,
          },
        },
        'sass-loader',
        {
          loader: `postcss-loader`, // translates CSS into CommonJS
          options: {
            syntax: 'postcss-scss',
            plugins: () => [
              require('postcss-import'),
              require('postcss-custom-media'),
              require('postcss-flexbugs-fixes'),
            ],
          },
        },
      ],
    }
    : {
      test: /\.s?css$/,
      exclude: /node_modules/,
      use: [
        {
          loader: `style-loader`,
          options: {
            singleton: false,
          },
        },
        {
          loader: `css-loader`,
          options: {
            importLoaders: 1,
            modules: true,
            sourceMap: true,
            localIdentName: `[local]_[hash:base64:3]`,
            camelCase: `dashes`,
          },
        },
        {
          loader: `postcss-loader`, // translates CSS into CommonJS
          options: {
            plugins: () => [
              require('postcss-nested'), //theme disable
              require('postcss-custom-media'),
              require('./postcss-css-variables/index.js'),
              require('postcss-flexbugs-fixes'),
            ],
          },
        },
        {
          loader: "sass-loader",
        },
        {
          loader: 'postcss-loader',
          options: {
            syntax: 'postcss-scss',
            plugins: () => [
              require('postcss-import'),
            ],
          },
        },
      ],
    }
  const entry = {block: './src/index.js'}
  return {
    target: 'node',
    entry,
    mode: 'development',
    output: {
      path: `/`,
      filename: `block.js`,
      libraryTarget: `commonjs`,
    },
    context: cwd,
    resolveLoader: {
      modules: [path.resolve(__dirname, `../../../node_modules`)],
    },
    resolve: {
      extensions: [`.js`],
      modules: [path.resolve(cwd, `node_modules`)],
    },
    module: {
      rules: [
        {
          test: /\.(js|jsx)$/,
          loader: `babel-loader`,
          exclude: /node_modules\/(?!wireframe-.*)/,
          options: {
            presets: [require(`babel-preset-env`), require(`babel-preset-react`)],
            plugins: [
              require(`babel-plugin-transform-class-properties`),
              require(`babel-plugin-transform-object-rest-spread`),
            ],
          },
        },
        {
          include: /node_modules/,
          test: /\.s?css$/,
          use: [{
            loader: `style-loader`,
            options: {
              singleton: false,
            },
          },
          {
            loader: `css-loader`,
            options: {
              modules: false,
            },
          },
          ],
        }, styleLoader,
      ],
    },
    externals: (context, request, callback) => {
      if (request.match(/style-loader\/lib\/addStyles\.js/)) {
        return callback(null, `commonjs style-loader/lib/addStyles.js`)
      }
      if (request.startsWith(`weblium/editor`)) {
        return callback(null, `weblium/editor`)
      }
      if (request.startsWith(`lodash`)) {
        return callback(null, `commonjs ${request}`)
      }
      if (request.startsWith(`react-router`)) {
        return callback(null, `commonjs ${request}`)
      }
      if (request.startsWith(`core-js`)) {
        return callback(null, `commonjs ${request}`)
      }
      return externalDeps[request] ? callback(null, externalDeps[request]) : callback()
    },
    plugins: [],
    devtool: false,
  }
}

let server
export const startServer = async ({port = 9999, uikitsList, blockUikitId, type, api}) => {
  const cwd = process.cwd()
  if (server) {
    return server
  }
  if (!realFs.statSync(path.join(cwd, `src`, `index.js`)).isFile()) {
    throw new Error(`Please create src folder with index.js entrypoint`)
  }

  const sandboxConfig = await getSandboxConfig(type)
  const blockConfig = await getBlockConfig(type)
  const compileBlock = webpack(blockConfig)
  compileBlock.outputFileSystem = fs
  await new Promise((resolve, reject) => {
    compileBlock.run((err, stats) => {
      if (err) {
        console.error(err)
      }
      if (stats.hasErrors()) {
        const stat = stats.toJson()
        process.stdout.write(stat.errors.join('\n----------------\n') + '\n')
        return reject(err)
      }
      resolve()
    })
  })

  const compileSandbox = webpack(sandboxConfig)

  loadTmpModules()

  sandboxVirtualModules.writeModule(
    `node_modules/weblium/sandbox-data.json`,
    JSON.stringify({uikitsList, blockUikitId}),
  )

  compileBlock.watch({ignored: /node_modules/}, (err, stats) => {
    if (err) {
      console.log(err.stack || err)
      if (err.details) {
        console.log(err.details)
      }
    }
    if (stats.hasErrors()) {
      const stat = stats.toJson()
      process.stdout.write(stat.errors.join('\n----------------\n') + '\n')
    }
    try {
      fs.statSync(`/block.js`).isFile()
      const blockModule = fs.readFileSync(`/block.js`).toString(`utf8`)
      sandboxVirtualModules.writeModule(`node_modules/block.js`, blockModule)
    } catch (e) {
      console.log(e)
    }
  })

  compileSandbox.hooks.afterCompile.tap('WebpackWatchPlugin', () => {
    if (fs.statSync('/block.js').isFile()) {
      const blockModule = fs.readFileSync('/block.js').toString('utf8')
      sandboxVirtualModules.writeModule('node_modules/block.js', blockModule)
    }
  })
  const app = new Koa()
  app.use(devMiddleware(compileSandbox, {
    noInfo: false,
    quiet: false,
    watchOptions: {
      aggregateTimeout: 300,
      poll: false,
      ignored: /node_modules/,
    },
    publicPath: `/`,
    stats: {
      colors: true,
    },
  }))
  app.use(hotMiddleware(compileSandbox, {}))

  app.use(serve(path.join(__dirname, `..`, `public`)))
  app.use(serve(path.join(cwd, `assets`)))
  server = http.createServer(app.callback())
  server.listen(port)

  if (api) {
    await new Promise((resolve, reject) => {
      console.log('initializing api')
      const io = require("socket.io")(server, {})

      io.on("error", reject)
      io.on("connection", (socket) => {
        resolve()
        socket.on("api", ({messageId, action, ...data}) => {
          const sendError = (error) => {
            console.log(`${action}:${messageId} => `, error)
            socket.emit("api", {messageId, error: {message: Error.message || "" + error}})
          }
          try {
            console.log(`${action}:${messageId} start`, data)
            const res = api[action](data)
            res.then((responseData) => {
              console.log(`${action}:${messageId}: end`)
              socket.emit("api", {...responseData, messageId})
            }).catch(sendError)
          } catch (e) {
            sendError(e)
          }
        })
      })
    })
  }

  port = server.address().port
  console.log(`server started on port ${port}`)
  return () => server.close()
}
