process.env.SANDBOX_BUILD = true
const path = require('path')
const fs = require('fs')
const webpack = require('../node_modules/webpack')
const SvgStore = require('webpack-svgstore-plugin')
const paths = require('../editor-next/config/paths')
const {
  provides,
  defines,
  optimization,
  mode,
  commonRules,
} = require('../editor-next/scripts/editor-configs')

const commonPlugins = [
  new webpack.optimize.LimitChunkCountPlugin({maxChunks: 1}),
  new webpack.ProvidePlugin(provides),
  new webpack.DefinePlugin(defines),
]

const sandboxMini = {
  mode,
  entry: paths.sandboxMini,
  output: {
    path: path.join(__dirname, 'app'),
    filename: 'sandbox-mini.js',
    libraryTarget: 'umd',
  },
  resolve: {
    modules: ['node_modules', paths.projectRoot, paths.clientRoot, paths.appSrc],
    extensions: ['.js', '.json', '.jsx'],
  },
  module: {
    rules: [
      ...commonRules,
      {
        test: /\.(js|jsx)$/,
        loader: 'babel-loader',
        include: [
          paths.appSrc,
        ],
        options: {
          presets: ['env', 'react'],
          plugins: [
            [
              'import',
              {
                libraryName: 'ui',
                style: true,
                libraryDirectory: 'src',
              },
            ],
            'universal-import',
            'syntax-dynamic-import',
            'transform-class-properties',
            'transform-object-rest-spread',
            'transform-decorators-legacy',
          ],
        },
      },
      {
        test: /\.s?css$/,
        include: [
          paths.appSrc
        ],
        use: [
          {
            loader: 'style-loader',
            options: {
              hmr: false,
            },
          },
          {
            loader: 'css-loader',
            options: {
              importLoaders: 1,
              modules: true,
              localIdentName: '[local]_[hash:base64:3]',
              camelCase: 'dashes',
            },
          },
          {
            loader: 'postcss-loader', // translates CSS into CommonJS
            options: {
              config: {
                path: paths.postcssConfig,
              },
            },
          },
        ],
      },
      {
        test: /\.s?css$/,
        include: new RegExp(`(node_modules|${paths.uiComponents})`),
        use: [
          {
            loader: 'style-loader',
            options: {
              hmr: false,
            },
          },
          {
            loader: 'css-loader',
            options: {
              importLoaders: 1,
            },
          },
          {
            loader: 'postcss-loader', // translates CSS into CommonJS
            options: {
              config: {
                path: paths.postcssConfig,
              },
            },
          },
        ],
      },
    ],
  },
  optimization,
  plugins: [...commonPlugins],
}

const sandboxFull = {
  mode,
  optimization,
  plugins: [
    ...commonPlugins,
    new SvgStore.Options({
      svgoOptions: {
        plugins: [
          {
            removeTitle: true,
          },
        ],
      },
    }),
  ],
  entry: ['babel-polyfill', paths.sandboxFull],
  output: {
    path: path.join(__dirname, 'app'),
    filename: 'sandbox-full.js',
    libraryTarget: 'umd',
  },
  resolve: {
    modules: ['node_modules', paths.projectRoot, paths.clientRoot, paths.appSrc],
    extensions: ['.js', '.json', '.jsx'],
  },
  module: {
    rules: [
      ...commonRules,
      {
        test: /\.(js|jsx)$/,
        loader: 'babel-loader',
        include: [
          paths.appSrc,
          /socket-api\/shared.js/,
          /socket-api\/attach.js/,
          paths.uiComponents,
          paths.mediaGallery,
          path.resolve(paths.clientRoot, 'shared'),
          new RegExp('transliteration/(?!node_modules)'),
        ],
        options: {
          presets: [[require.resolve('babel-preset-env'), {targets: {browsers: ['last 2 versions']}}], require.resolve('babel-preset-react')],
          babelrc: false,
          plugins: [
            require.resolve('babel-plugin-universal-import'),
            require.resolve('babel-plugin-syntax-dynamic-import'),
            [
              require.resolve('babel-plugin-module-resolver'),
              {
                alias: {
                  '^services/api': () => path.join(paths.appSrc, 'services/stub'),
                  'services/store': () => path.join(paths.appSrc, 'services/store-stub'),
                },
              },
            ],
            [
              require.resolve('babel-plugin-import'),
              {
                libraryName: 'ui',
                style: true,
                libraryDirectory: 'src',
              },
            ],
            require.resolve('babel-plugin-transform-class-properties'),
            require.resolve('babel-plugin-transform-object-rest-spread'),
            require.resolve('babel-plugin-transform-decorators-legacy')
          ],
        },
      },
      {
        test: /\.s?css$/,
        include: [paths.appSrc, paths.mediaGallery, path.resolve(paths.clientRoot, 'shared')],
        use: [
          {
            loader: 'style-loader',
            options: {
              hmr: false,
            },
          },
          {
            loader: 'css-loader',
            options: {
              importLoaders: 1,
              modules: true,
              localIdentName: '[local]_[hash:base64:3]',
              camelCase: 'dashes',
            },
          },
          {
            loader: 'postcss-loader', // translates CSS into CommonJS
            options: {
              config: {
                path: paths.postcssConfig,
              },
            },
          },
        ],
      },
      {
        test: /\.s?css$/,
        include: new RegExp(`(node_modules|${paths.uiComponents})`),
        use: [
          {
            loader: 'style-loader',
            options: {
              hmr: false,
            },
          },
          {
            loader: 'css-loader',
            options: {
              importLoaders: 1,
            },
          },
          {
            loader: 'postcss-loader', // translates CSS into CommonJS
            options: {
              config: {
                path: paths.postcssConfig,
              },
            },
          },
        ],
      },
    ],
  },
  devtool: false,
}

module.exports = async () => {
  const {version: editorVersion} = require('../editor-next/package.json')
  const config = [sandboxMini, sandboxFull]
  const sandboxFullPath = path.join(__dirname, 'app', 'sandbox-full.js')
  const sandboxMiniPath = path.join(__dirname, 'app', 'sandbox-mini.js')
  try {fs.unlinkSync(sandboxFullPath)} catch (err) {}
  try {fs.unlinkSync(sandboxMiniPath)} catch (err) {}
  const compiler = webpack(config)
  try {
    await new Promise((resolve, reject) => compiler.run((err, stats) => {
      if (err) {
        console.log(err)
        return reject(err)
      }
      if (stats.hasErrors()) {
        const stat = stats.toJson()
        process.stdout.write(stat.errors.join('\n----------------\n') + '\n')
        return reject(stats)
      }
      resolve()
    }))
    const oldSpritePath = path.join(__dirname, 'app', 'sprite.svg')
    const newSpritePath = path.join(__dirname, 'app', 'sandbox', 'public', 'sprite.svg')
    fs.renameSync(oldSpritePath, newSpritePath)
    fs.writeFileSync(path.join(__dirname, 'app', 'editor-config.json'), JSON.stringify({editorVersion}))
    console.log('done: writing file')
  } catch (e) {
    console.log(e)
    throw e
  }
}

if(!module.parent) {
  module.exports().then(() => {
    process.exit(0)
  })
}



