const assert = require("assert");
const fetch = require("node-fetch");
const { flow } = require("lodash");
/**
 * Parse gs ref
 * @param {string} ref
 * @returns {Object} {bucketName, path}
 */
function parse(ref) {
  if (!ref || !ref.match) return null;
  const matches = ref.match(
    /^gs:\/\/(\w[\w.-]+)\/((?:\w[\w.-]*)(?:\/\w[\w.-]*)*)$/
  );
  if (!matches) return null;
  const [, bucketName, path] = matches;
  return { bucketName, path };
}

/**
 * Returns url by gs ref
 * @param {string} ref
 * @returns {string} Url from storage.googleapis.com
 */
function ref2url(ref) {
  const refo = parse(ref);
  assert(refo, "Invalid reference");
  const { path, bucketName } = refo;
  return bucketName.indexOf(".") >= 0
    ? `https://${bucketName}/${path}`
    : `https://storage.googleapis.com/${bucketName}/${path}`;
}

const getContent = async contentRef => {
  if (!contentRef) {
    return null;
  }
  const response = await fetch(ref2url(contentRef));
  if (response.status === 200) {
    const content = await response.json();
    return content;
  }

  return {};
};

const getBlockContent = block => {
  if (block.contentCache) {
    try {
      const content = JSON.parse(block.contentCache);
      return Promise.resolve(content);
    } catch (e) {
      console.log(e);
      return Promise.resolve(null);
    }
  }
  return block.contentRef
    ? getContent(block.contentRef)
    : Promise.resolve(null);
};

const composeMigration = (migration, blockVersion, baseVersion) => {
  const exports = {};
  const res = Function(
    "exports,module",
    `var _ = module._;${migration};\nreturn module.exports`
  )(exports, { exports, _: require('lodash/fp')});

  if (!res || typeof res != "object" || Array.isArray(res))
    throw new Error(`Invalid migration data: ${res}`);

  const sortedVersions = Object.keys(res)
    .filter(a => a.match(/^(\d+)\.(\d+)\.(\d+)$/))
    .map(v => v.split("."))
    .filter(([major]) => major > blockVersion.major && major <= baseVersion.major) //skip migration lower current version
    .sort((a, b) => {
      const [aMajor] = a;
      const [bMajor] = b;

      return aMajor >= bMajor ? 1 : -1;
    }); //sort by major version

  console.log(`available migrations: \n${sortedVersions.map(v => `* ${v.join('.')}`).join('\n')}`)

  return flow(sortedVersions.map(v => res[v.join(".")]));
};


const separator = Array(50).fill('-').join('').concat('\n')

const colorType = {
  msg: "\x1b[32m%s\x1b[0m",
  error: "\x1b[31m%s\x1b[0m",
}

const printByType = type => (string, withSeparator = true) => console.log(colorType[type], `\n${string}\n${withSeparator ? separator : ''}`)

const printMessage = printByType('msg')
const printError = printByType('error')

const migrateBlock = async (api, { location = process.cwd(), id, test = false}) => {
  try {
    const {
      data: { base, ...block }
    } = await api.block.get({ id });

    if (!base) {
      printMessage(`block ${id} doesn't have base to update`);
      return;
    }
    const { data: baseBlock } = await api.block.get({ id: base._id });

    if (!baseBlock.migrationRef) {
      printMessage(`base block ${base._id} doesn't have migration`);
      return;
    }

    if (block.version.major < base.version.major) {
      const migrationUrl = ref2url(baseBlock.migrationRef);
      const migrationCode = await fetch(migrationUrl).then(res => res.text());
      const content = await getBlockContent(block);
      try {
        const migrationsCompose = composeMigration(migrationCode, block.version, base.version)
        const {content: newContent, modifier: newModifier} = migrationsCompose({modifier: JSON.parse(block.modifier), content})

        if (test) {
          printMessage('MIGRATION RESULT:', false)
          printMessage('CONTENT:')
          console.log(JSON.stringify(newContent, null, 2))
          printMessage('MODIFIER:', JSON.stringify(newModifier, null, 2))
          console.log(JSON.stringify(newContent, null, 2))
        } else {
          const toUpdateData = {}

          newContent && (toUpdateData.content = newContent)
          newModifier && (toUpdateData.modifier = JSON.stringify(newModifier))
          const rerollRes = await api.block.reroll({id, source: base._id, data: toUpdateData})
          
          printMessage('REROLL RESIULT')
          console.log(JSON.stringify(rerollRes, null, 2))
        }

      } catch (error) {
        printError(`migration went wrong: ${error}`)
      }

    } else {
      printMessage(`the block ${id} is already up to date`);
    }
  } catch (error) {
    printError(error);
  }
};

module.exports = migrateBlock;
